@@ -54,9 +54,6 @@ class ArrayVarAddressReplacer: public ASR::BaseExprReplacer<ArrayVarAddressRepla
     void replace_ArrayItem(ASR::ArrayItem_t* /*x*/) {
     }
 
-    void replace_IntrinsicArrayFunction(ASR::IntrinsicArrayFunction_t* /*x*/) {
-    }
-
     void replace_FunctionCall(ASR::FunctionCall_t* x) {
         if( !ASRUtils::is_elemental(x->m_name) ) {
             return ;
@@ -95,7 +92,7 @@ class ArrayVarAddressCollector: public ASR::CallReplacerOnExpressionsVisitor<Arr
     }
 
     void visit_SubroutineCall(const ASR::SubroutineCall_t& x) {
-        if( !PassUtils::is_elemental(x.m_name) ) {
+        if( !ASRUtils::is_elemental(x.m_name) ) {
             return ;
         }
     }
@@ -112,16 +109,6 @@ class FixTypeVisitor: public ASR::CallReplacerOnExpressionsVisitor<FixTypeVisito
         (void)al_;      // Explicitly mark the parameter as unused
     }
 
-    void visit_StructType(const ASR::StructType_t& x) {
-        std::string derived_type_name = ASRUtils::symbol_name(x.m_derived_type);
-        if( x.m_derived_type == current_scope->resolve_symbol(derived_type_name) ) {
-            return ;
-        }
-
-        ASR::StructType_t& xx = const_cast<ASR::StructType_t&>(x);
-        xx.m_derived_type = current_scope->resolve_symbol(derived_type_name);
-    }
-
     void visit_Cast(const ASR::Cast_t& x) {
         ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_Cast(x);
         ASR::Cast_t& xx = const_cast<ASR::Cast_t&>(x);
@@ -143,11 +130,12 @@ class FixTypeVisitor: public ASR::CallReplacerOnExpressionsVisitor<FixTypeVisito
 
     void visit_FunctionCall(const ASR::FunctionCall_t& x) {
         ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_FunctionCall(x);
-        if( !PassUtils::is_elemental(x.m_name) ) {
+        if( !ASRUtils::is_elemental(x.m_name) ) {
             return ;
         }
         ASR::FunctionCall_t& xx = const_cast<ASR::FunctionCall_t&>(x);
-        if( !ASRUtils::is_array(ASRUtils::expr_type(x.m_args[0].m_value)) ) {
+        if( (x.m_dt && !ASRUtils::is_array(ASRUtils::expr_type(x.m_dt))) ||
+            !ASRUtils::is_array(ASRUtils::expr_type(x.m_args[0].m_value)) ) {
             xx.m_type = ASRUtils::extract_type(xx.m_type);
             xx.m_value = nullptr;
         }
@@ -168,6 +156,11 @@ class FixTypeVisitor: public ASR::CallReplacerOnExpressionsVisitor<FixTypeVisito
         visit_ArrayOp(x);
     }
 
+    void visit_ComplexBinOp(const ASR::ComplexBinOp_t& x) {
+        ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_ComplexBinOp(x);
+        visit_ArrayOp(x);
+    }
+
     void visit_RealCompare(const ASR::RealCompare_t& x) {
         ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_RealCompare(x);
         visit_ArrayOp(x);
@@ -188,6 +181,11 @@ class FixTypeVisitor: public ASR::CallReplacerOnExpressionsVisitor<FixTypeVisito
         visit_ArrayOp(x);
     }
 
+    void visit_StringConcat(const ASR::StringConcat_t& x){
+        ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_StringConcat(x);
+        visit_ArrayOp(x);
+    }
+
     void visit_LogicalBinOp(const ASR::LogicalBinOp_t& x) {
         ASR::CallReplacerOnExpressionsVisitor<FixTypeVisitor>::visit_LogicalBinOp(x);
         visit_ArrayOp(x);
@@ -250,26 +248,53 @@ class ReplaceArrayOp: public ASR::BaseExprReplacer<ReplaceArrayOp> {
 
     void replace_ArrayConstant(ASR::ArrayConstant_t* x) {
         remove_original_stmt_if_size_0(x->m_type)
+        if (result_expr == nullptr) {
+            return;
+        }
         pass_result.reserve(al, x->m_n_data);
         const Location& loc = x->base.base.loc;
-        LCOMPILERS_ASSERT(result_expr != nullptr);
         ASR::ttype_t* result_type = ASRUtils::expr_type(result_expr);
         ASR::ttype_t* result_element_type = ASRUtils::extract_type(result_type);
-        for( int64_t i = 0; i < ASRUtils::get_fixed_size_of_array(x->m_type); i++ ) {
+
+        ASR::dimension_t* m_dims = nullptr;
+        size_t n_dims = ASRUtils::extract_dimensions_from_ttype(x->m_type, m_dims);
+
+        int64_t size = ASRUtils::get_fixed_size_of_array(x->m_type);
+        int repeat = 1;
+        std::vector<std::vector<int64_t>> rank_indexes(n_dims, std::vector<int64_t>(size));
+        for (size_t i = 0; i < n_dims; i++) {
+            int64_t length = 0, start = 1, ubound;
+            ASRUtils::extract_value(m_dims[i].m_length, length);
+            ASRUtils::extract_value(m_dims[i].m_start, start);
+            ubound = length + start - 1;
+            int64_t c = 0;
+            while (c != size) {
+                for (int64_t j = start; j <= ubound; j++) {
+                    for (int64_t k = 1; k <= repeat; k++) {
+                        rank_indexes[i][c++] = j;
+                    }
+                }
+            }
+            repeat *= length;
+        }
+
+        for (int64_t i = 0; i < size; i++) {
             ASR::expr_t* x_i = ASRUtils::fetch_ArrayConstant_value(al, x, i);
             Vec<ASR::array_index_t> array_index_args;
-            array_index_args.reserve(al, 1);
-            ASR::array_index_t array_index_arg;
-            array_index_arg.loc = loc;
-            array_index_arg.m_left = nullptr;
-            array_index_arg.m_right = make_ConstantWithKind(
-                make_IntegerConstant_t, make_Integer_t, i + 1, 4, loc);
-            array_index_arg.m_step = nullptr;
-            array_index_args.push_back(al, array_index_arg);
+            array_index_args.reserve(al, n_dims);
+            for (size_t j = 0; j < n_dims; j++) {
+                ASR::array_index_t array_index_arg;
+                array_index_arg.loc = loc;
+                array_index_arg.m_left = nullptr;
+                array_index_arg.m_right = make_ConstantWithKind(
+                    make_IntegerConstant_t, make_Integer_t, rank_indexes[j][i], 4, loc);
+                array_index_arg.m_step = nullptr;
+                array_index_args.push_back(al, array_index_arg);
+            }
             ASR::expr_t* y_i = ASRUtils::EXPR(ASRUtils::make_ArrayItem_t_util(al, loc,
                 result_expr, array_index_args.p, array_index_args.size(),
                 result_element_type, ASR::arraystorageType::ColMajor, nullptr));
-            pass_result.push_back(al, ASRUtils::STMT(ASRUtils::make_Assignment_t_util(al, loc, y_i, x_i, nullptr, false)));
+            pass_result.push_back(al, ASRUtils::STMT(ASRUtils::make_Assignment_t_util(al, loc, y_i, x_i, nullptr, false, false)));
         }
     }
 
@@ -313,9 +338,11 @@ class ReplaceArrayOp: public ASR::BaseExprReplacer<ReplaceArrayOp> {
 
         remove_original_stmt_if_size_0(arr_type)
 
+        if (result_expr == nullptr) {
+            return;
+        }
         pass_result.reserve(al, x->n_args);
         const Location& loc = x->base.base.loc;
-        LCOMPILERS_ASSERT(result_expr != nullptr);
 
         ASR::ttype_t* result_type = ASRUtils::expr_type(result_expr);
         ASRUtils::ExprStmtDuplicator duplicator(al);
@@ -326,6 +353,10 @@ class ReplaceArrayOp: public ASR::BaseExprReplacer<ReplaceArrayOp> {
         fix_type_visitor.current_scope = current_scope;
         fix_type_visitor.visit_ttype(*result_element_type);
 
+        ASRUtils::ASRBuilder builder(al, loc);
+        ASR::dimension_t* m_dims = nullptr;
+        ASRUtils::extract_dimensions_from_ttype(arr_type, m_dims);
+
         for( int64_t i = 0; i < ASRUtils::get_fixed_size_of_array(arr_type); i++ ) {
             ASR::expr_t* x_i = nullptr;
             if( x->m_value ) {
@@ -339,14 +370,14 @@ class ReplaceArrayOp: public ASR::BaseExprReplacer<ReplaceArrayOp> {
             ASR::array_index_t array_index_arg;
             array_index_arg.loc = loc;
             array_index_arg.m_left = nullptr;
-            array_index_arg.m_right = make_ConstantWithKind(
-                make_IntegerConstant_t, make_Integer_t, i + 1, 4, loc);
+            array_index_arg.m_right = builder.Add(m_dims[0].m_start, make_ConstantWithKind(
+                make_IntegerConstant_t, make_Integer_t, i, 4, loc));
             array_index_arg.m_step = nullptr;
             array_index_args.push_back(al, array_index_arg);
             ASR::expr_t* y_i = ASRUtils::EXPR(ASRUtils::make_ArrayItem_t_util(al, loc,
                 result_expr, array_index_args.p, array_index_args.size(),
                 result_element_type, ASR::arraystorageType::ColMajor, nullptr));
-            pass_result.push_back(al, ASRUtils::STMT(ASRUtils::make_Assignment_t_util(al, loc, y_i, x_i, nullptr, false)));
+            pass_result.push_back(al, ASRUtils::STMT(ASRUtils::make_Assignment_t_util(al, loc, y_i, x_i, nullptr, false, false)));
         }
     }
 
@@ -360,6 +391,63 @@ ASR::expr_t* at(Vec<ASR::expr_t*>& vec, int64_t index) {
     return vec[index];
 }
 
+// Collect the array expressions which should be of the same size
+class CollectComponentsFromElementalExpr: public ASR::BaseWalkVisitor<CollectComponentsFromElementalExpr> {
+private:
+        Allocator& al;
+        Vec<ASR::expr_t*>& vars;
+public:
+
+        CollectComponentsFromElementalExpr(Allocator& al_, Vec<ASR::expr_t*>& vars_) :
+        al(al_), vars(vars_) {}
+
+        void push_expr_if_array(ASR::expr_t *v) {
+            if (ASRUtils::is_array(ASRUtils::expr_type(v))) {
+                vars.push_back(al, v);
+            }
+        }
+
+        // Don't go inside these
+        void visit_ttype(const ASR::ttype_t &) {}
+        void visit_ArraySection(const ASR::ArraySection_t&) {}
+        void visit_ArrayItem(const ASR::ArrayItem_t&) {}
+        void visit_ArraySize(const ASR::ArraySize_t&) {}
+        void visit_ArrayReshape(const ASR::ArrayReshape_t&) {}
+        void visit_ArrayBound(const ASR::ArrayBound_t&) {}
+
+        void visit_Var(const ASR::Var_t& x) {
+            ASR::Var_t *xx = const_cast<ASR::Var_t*>(&x);
+            push_expr_if_array((ASR::expr_t *)xx);
+        }
+
+        void visit_ArrayPhysicalCast(const ASR::ArrayPhysicalCast_t& x) {
+            ASR::ArrayPhysicalCast_t *xx = const_cast<ASR::ArrayPhysicalCast_t*>(&x);
+            push_expr_if_array((ASR::expr_t *)xx);
+        }
+
+        void visit_StructInstanceMember(const ASR::StructInstanceMember_t& x) {
+            ASR::StructInstanceMember_t *xx = const_cast<ASR::StructInstanceMember_t*>(&x);
+            push_expr_if_array((ASR::expr_t *)xx);
+        }
+
+        void visit_BitCast(const ASR::BitCast_t& x) {
+            ASR::BitCast_t *xx = const_cast<ASR::BitCast_t*>(&x);
+            push_expr_if_array((ASR::expr_t *)xx);
+        }
+
+        void visit_ArrayConstant(const ASR::ArrayConstant_t& x) {
+            ASR::ArrayConstant_t *xx = const_cast<ASR::ArrayConstant_t*>(&x);
+            push_expr_if_array((ASR::expr_t *)xx);
+        }
+
+        // Only go inside the FunctionCall if the Function is elemental
+        void visit_FunctionCall(const ASR::FunctionCall_t& x) {
+            if (ASRUtils::is_elemental(x.m_name)) {
+                ASR::BaseWalkVisitor<CollectComponentsFromElementalExpr>::visit_FunctionCall(x);
+            }
+        }
+};
+
 class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisitor> {
     private:
 
@@ -368,7 +456,9 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
     Vec<ASR::stmt_t*> pass_result;
     Vec<ASR::stmt_t*>* parent_body;
     bool realloc_lhs;
+    bool bounds_checking;
     bool remove_original_stmt;
+    inline static std::set<const ASR::Assignment_t*> debug_inserted;
 
     public:
 
@@ -378,9 +468,9 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         replacer.replace_expr(*current_expr);
     }
 
-    ArrayOpVisitor(Allocator& al_, bool realloc_lhs_):
+    ArrayOpVisitor(Allocator& al_, bool realloc_lhs_, bool bounds_checking_):
         al(al_), replacer(al, pass_result, remove_original_stmt),
-        parent_body(nullptr), realloc_lhs(realloc_lhs_),
+        parent_body(nullptr), realloc_lhs(realloc_lhs_), bounds_checking(bounds_checking_),
         remove_original_stmt(false) {
         pass_result.n = 0;
         pass_result.reserve(al, 0);
@@ -444,33 +534,41 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
             if( i == var_with_maxrank ) {
                 continue;
             }
-            ASR::expr_t* index_var = at(var2indices[i], loop_depth);
+            // Skip variables with lower rank than the current loop depth
+            if( loop_depth >= static_cast<int64_t>(var2indices[i].n) ) {
+                continue;
+            }
+            ASR::expr_t* index_var = var2indices[i].p[loop_depth];
             if( index_var == nullptr ) {
                 continue;
             }
             ASR::expr_t* plus_one = ASRUtils::EXPR(ASR::make_IntegerBinOp_t(al, loc, index_var,
                 ASR::binopType::Add, step, ASRUtils::expr_type(index_var), nullptr));
             ASR::stmt_t* increment = ASRUtils::STMT(ASRUtils::make_Assignment_t_util(
-                al, loc, index_var, plus_one, nullptr, false));
+                al, loc, index_var, plus_one, nullptr, false, false));
             do_loop_body.push_back(al, increment);
         }
     }
 
     void set_index_variables(std::unordered_map<size_t, Vec<ASR::expr_t*>>& var2indices,
-                             Vec<ASR::expr_t*>& vars_expr, size_t var_with_maxrank, size_t max_rank,
+                             Vec<ASR::expr_t*>& vars_expr, size_t var_with_maxrank,
                              int64_t loop_depth, Vec<ASR::stmt_t*>& dest_vec, const Location& loc) {
         for( size_t i = 0; i < var2indices.size(); i++ ) {
             if( i == var_with_maxrank ) {
                 continue;
             }
-            ASR::expr_t* index_var = at(var2indices[i], loop_depth);
+            // Skip variables with lower rank than the current loop depth
+            if( loop_depth >= static_cast<int64_t>(var2indices[i].n) ) {
+                continue;
+            }
+            ASR::expr_t* index_var = var2indices[i].p[loop_depth];
             if( index_var == nullptr ) {
                 continue;
             }
             ASR::expr_t* lbound = PassUtils::get_bound(vars_expr[i],
-                loop_depth + max_rank + 1, "lbound", al);
+                loop_depth + 1, "lbound", al);
             ASR::stmt_t* set_index_var = ASRUtils::STMT(ASRUtils::make_Assignment_t_util(
-                al, loc, index_var, lbound, nullptr, false));
+                al, loc, index_var, lbound, nullptr, false, false));
             dest_vec.push_back(al, set_index_var);
         }
     }
@@ -481,17 +579,21 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
 
     void set_index_variables(std::unordered_map<size_t, Vec<ASR::expr_t*>>& var2indices,
                              std::unordered_map<ASR::expr_t*, std::pair<ASR::expr_t*, IndexType>>& index2var,
-                             size_t var_with_maxrank, size_t max_rank, int64_t loop_depth,
+                             size_t var_with_maxrank, int64_t loop_depth,
                              Vec<ASR::stmt_t*>& dest_vec, const Location& loc) {
         for( size_t i = 0; i < var2indices.size(); i++ ) {
             if( i == var_with_maxrank ) {
                 continue;
             }
-            ASR::expr_t* index_var = at(var2indices[i], loop_depth);
+            // Skip variables with lower rank than the current loop depth
+            if( loop_depth >= static_cast<int64_t>(var2indices[i].n) ) {
+                continue;
+            }
+            ASR::expr_t* index_var = var2indices[i].p[loop_depth];
             if( index_var == nullptr ) {
                 continue;
             }
-            size_t bound_dim = loop_depth + max_rank + 1;
+            size_t bound_dim = loop_depth + 1;
             if( index2var[index_var].second == IndexType::ArrayIndex ) {
                 bound_dim = 1;
             }
@@ -503,7 +605,7 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
                     index2var[index_var].first, bound_dim, "lbound", al);
             }
             ASR::stmt_t* set_index_var = ASRUtils::STMT(ASRUtils::make_Assignment_t_util(
-                al, loc, index_var, lbound, nullptr, false));
+                al, loc, index_var, lbound, nullptr, false, false));
             dest_vec.push_back(al, set_index_var);
         }
     }
@@ -598,7 +700,7 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         Vec<ASR::array_index_t> rhs_array_indices_args; rhs_array_indices_args.reserve(al, 1);
         int n_array_indices_args = -1;
         int temp_n = -1;
-        size_t do_loop_depth = var_rank;
+        size_t do_loop_depth = 0;
         if( is_target_array ) {
             vars_expr.push_back(al, ASRUtils::extract_array_variable(target));
             ASRUtils::extract_array_indices(target, al, array_indices_args, n_array_indices_args);
@@ -633,7 +735,7 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
                     ASR::abiType::Source, ASR::accessType::Public, ASR::presenceType::Required, false));
                 current_scope->add_symbol(index_var_name, index);
                 ASR::expr_t* index_expr = ASRUtils::EXPR(ASR::make_Var_t(al, loc, index));
-                if ((i == offset_for_array_indices - 1) && is_value_array &&
+                if ((i == offset_for_array_indices - 1) && is_value_array && j < rhs_array_indices_args.size() &&
                         rhs_array_indices_args[j].m_left != nullptr) {
                     index2var[index_expr] = std::make_pair(rhs_array_indices_args[j].m_left, IndexType::ScalarIndex);
                 } else {
@@ -664,17 +766,18 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
 
         ASR::do_loop_head_t do_loop_head;
         do_loop_head.loc = loc;
-        do_loop_head.m_v = at(var2indices[var_with_maxrank], -1);
-        size_t bound_dim = do_loop_depth;
+        do_loop_head.m_v = var2indices[var_with_maxrank].p[0];
+        size_t bound_dim = do_loop_depth + 1;
         if( index2var[do_loop_head.m_v].second == IndexType::ArrayIndex ) {
             bound_dim = 1;
         }
-        if( n_array_indices_args > -1 && array_indices_args[n_array_indices_args].m_right != nullptr &&
-                array_indices_args[n_array_indices_args].m_left != nullptr &&
-                array_indices_args[n_array_indices_args].m_step != nullptr) {
-            do_loop_head.m_start = array_indices_args[n_array_indices_args].m_left;
-            do_loop_head.m_end = array_indices_args[n_array_indices_args].m_right;
-            do_loop_head.m_increment = array_indices_args[n_array_indices_args].m_step;
+        int64_t array_indices_args_i = 0;
+        if( n_array_indices_args > -1 && array_indices_args[array_indices_args_i].m_right != nullptr &&
+                array_indices_args[array_indices_args_i].m_left != nullptr &&
+                array_indices_args[array_indices_args_i].m_step != nullptr) {
+            do_loop_head.m_start = array_indices_args[array_indices_args_i].m_left;
+            do_loop_head.m_end = array_indices_args[array_indices_args_i].m_right;
+            do_loop_head.m_increment = array_indices_args[array_indices_args_i].m_step;
         } else {
             do_loop_head.m_start = PassUtils::get_bound(
                 index2var[do_loop_head.m_v].first, bound_dim, "lbound", al);
@@ -685,36 +788,36 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         Vec<ASR::stmt_t*> parent_do_loop_body; parent_do_loop_body.reserve(al, 1);
         Vec<ASR::stmt_t*> do_loop_body; do_loop_body.reserve(al, 1);
         set_index_variables(var2indices, index2var, var_with_maxrank,
-                            var_rank, -1, parent_do_loop_body, loc);
+                            0, parent_do_loop_body, loc);
         do_loop_body.push_back(al, const_cast<ASR::stmt_t*>(&(x.base)));
-        increment_index_variables(var2indices, var_with_maxrank, -1,
+        increment_index_variables(var2indices, var_with_maxrank, 0,
                                   do_loop_body, loc);
         ASR::stmt_t* do_loop = ASRUtils::STMT(ASR::make_DoLoop_t(al, loc, nullptr,
             do_loop_head, do_loop_body.p, do_loop_body.size(), nullptr, 0));
-        do_loop_depth--;
-        n_array_indices_args--;
+        do_loop_depth += 1;
+        array_indices_args_i += 1;
         parent_do_loop_body.push_back(al, do_loop);
         do_loop_body.from_pointer_n_copy(al, parent_do_loop_body.p, parent_do_loop_body.size());
         parent_do_loop_body.reserve(al, 1);
 
-        for( int64_t i = -2; i >= -static_cast<int64_t>(var_rank); i-- ) {
+        for( int64_t i = 1; i < static_cast<int64_t>(var_rank); i++ ) {
             set_index_variables(var2indices, index2var, var_with_maxrank,
-                                var_rank, i, parent_do_loop_body, loc);
+                                i, parent_do_loop_body, loc);
             increment_index_variables(var2indices, var_with_maxrank, i,
                                       do_loop_body, loc);
             ASR::do_loop_head_t do_loop_head;
             do_loop_head.loc = loc;
-            do_loop_head.m_v = at(var2indices[var_with_maxrank], i);
-            bound_dim = do_loop_depth;
+            do_loop_head.m_v = var2indices[var_with_maxrank].p[i];
+            bound_dim = do_loop_depth + 1;
             if( index2var[do_loop_head.m_v].second == IndexType::ArrayIndex ) {
                 bound_dim = 1;
             }
-            if( n_array_indices_args > -1 && array_indices_args[n_array_indices_args].m_right != nullptr &&
-                    array_indices_args[n_array_indices_args].m_left != nullptr &&
-                    array_indices_args[n_array_indices_args].m_step != nullptr) {
-                do_loop_head.m_start = array_indices_args[n_array_indices_args].m_left;
-                do_loop_head.m_end = array_indices_args[n_array_indices_args].m_right;
-                do_loop_head.m_increment = array_indices_args[n_array_indices_args].m_step;
+            if( n_array_indices_args > -1 && array_indices_args[array_indices_args_i].m_right != nullptr &&
+                    array_indices_args[array_indices_args_i].m_left != nullptr &&
+                    array_indices_args[array_indices_args_i].m_step != nullptr) {
+                do_loop_head.m_start = array_indices_args[array_indices_args_i].m_left;
+                do_loop_head.m_end = array_indices_args[array_indices_args_i].m_right;
+                do_loop_head.m_increment = array_indices_args[array_indices_args_i].m_step;
             } else {
                 do_loop_head.m_start = PassUtils::get_bound(
                     index2var[do_loop_head.m_v].first, bound_dim, "lbound", al);
@@ -724,8 +827,8 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
             }
             ASR::stmt_t* do_loop = ASRUtils::STMT(ASR::make_DoLoop_t(al, loc, nullptr,
                 do_loop_head, do_loop_body.p, do_loop_body.size(), nullptr, 0));
-            do_loop_depth--;
-            n_array_indices_args--;
+            do_loop_depth++;
+            array_indices_args_i += 1;
             parent_do_loop_body.push_back(al, do_loop);
             do_loop_body.from_pointer_n_copy(al, parent_do_loop_body.p, parent_do_loop_body.size());
             parent_do_loop_body.reserve(al, 1);
@@ -807,18 +910,18 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
 
         ASR::do_loop_head_t do_loop_head;
         do_loop_head.loc = loc;
-        do_loop_head.m_v = at(var2indices[var_with_maxrank], -1);
+        do_loop_head.m_v = var2indices[var_with_maxrank].p[0];
         do_loop_head.m_start = PassUtils::get_bound(vars_expr[var_with_maxrank],
-            var_ranks[var_with_maxrank], "lbound", al);
+            1, "lbound", al);
         do_loop_head.m_end = PassUtils::get_bound(vars_expr[var_with_maxrank],
-            var_ranks[var_with_maxrank], "ubound", al);
+            1, "ubound", al);
         do_loop_head.m_increment = nullptr;
         Vec<ASR::stmt_t*> parent_do_loop_body; parent_do_loop_body.reserve(al, 1);
         Vec<ASR::stmt_t*> do_loop_body; do_loop_body.reserve(al, 1);
         set_index_variables(var2indices, vars_expr, var_with_maxrank,
-                            var_ranks[var_with_maxrank], -1, parent_do_loop_body, loc);
+                            0, parent_do_loop_body, loc);
         do_loop_body.push_back(al, const_cast<ASR::stmt_t*>(&(x.base)));
-        increment_index_variables(var2indices, var_with_maxrank, -1,
+        increment_index_variables(var2indices, var_with_maxrank, 0,
                                   do_loop_body, loc);
         ASR::stmt_t* do_loop = ASRUtils::STMT(ASR::make_DoLoop_t(al, loc, nullptr,
             do_loop_head, do_loop_body.p, do_loop_body.size(), nullptr, 0));
@@ -826,18 +929,18 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         do_loop_body.from_pointer_n_copy(al, parent_do_loop_body.p, parent_do_loop_body.size());
         parent_do_loop_body.reserve(al, 1);
 
-        for( int64_t i = -2; i >= -static_cast<int64_t>(var_ranks[var_with_maxrank]); i-- ) {
+        for( int64_t i = 1; i < static_cast<int64_t>(var_ranks[var_with_maxrank]); i++ ) {
             set_index_variables(var2indices, vars_expr, var_with_maxrank,
-                                var_ranks[var_with_maxrank], i, parent_do_loop_body, loc);
+                                i, parent_do_loop_body, loc);
             increment_index_variables(var2indices, var_with_maxrank, i,
                                       do_loop_body, loc);
             ASR::do_loop_head_t do_loop_head;
             do_loop_head.loc = loc;
-            do_loop_head.m_v = at(var2indices[var_with_maxrank], i);
-            do_loop_head.m_start = PassUtils::get_bound(vars_expr[var_with_maxrank],
-                var_ranks[var_with_maxrank] + i + 1, "lbound", al);
-            do_loop_head.m_end = PassUtils::get_bound(vars_expr[var_with_maxrank],
-                var_ranks[var_with_maxrank] + i + 1, "ubound", al);
+            do_loop_head.m_v = var2indices[var_with_maxrank].p[i];
+            do_loop_head.m_start = PassUtils::get_bound(
+                vars_expr[var_with_maxrank], i + 1, "lbound", al);
+            do_loop_head.m_end = PassUtils::get_bound(
+                vars_expr[var_with_maxrank], i + 1, "ubound", al);
             do_loop_head.m_increment = nullptr;
             ASR::stmt_t* do_loop = ASRUtils::STMT(ASR::make_DoLoop_t(al, loc, nullptr,
                 do_loop_head, do_loop_body.p, do_loop_body.size(), nullptr, 0));
@@ -853,27 +956,22 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
 
     void insert_realloc_for_target(ASR::expr_t* target, ASR::expr_t* value, Vec<ASR::expr_t**>& vars) {
         ASR::ttype_t* target_type = ASRUtils::expr_type(target);
-        if( (realloc_lhs == false || !ASRUtils::is_allocatable(target_type) || vars.size() == 1) &&
-            !(ASR::is_a<ASR::Var_t>(*value) && ASR::is_a<ASR::Var_t>(*target) &&
-              ASRUtils::is_allocatable(target_type)) ) {
+        bool array_copy = ASR::is_a<ASR::Var_t>(*value) && ASR::is_a<ASR::Var_t>(*target);
+        if (!realloc_lhs) {
+            return;
+        }
+        if( (!ASRUtils::is_allocatable(target_type) || vars.size() == 1) &&
+            !(array_copy && ASRUtils::is_allocatable(target_type)) ) {
             return ;
         }
 
-        // First element in vars is target itself
-        ASR::expr_t* realloc_var = nullptr;
-        size_t target_rank = ASRUtils::extract_n_dims_from_ttype(target_type);
-        for( size_t i = 1; i < vars.size(); i++ ) {
-            size_t var_rank = ASRUtils::extract_n_dims_from_ttype(
-                ASRUtils::expr_type(*vars[i]));
-            if( target_rank == var_rank ) {
-                realloc_var = *vars[i];
-                break ;
-            }
-        }
+        ASRUtils::ExprStmtDuplicator d(al);
+        ASR::expr_t* realloc_var = d.duplicate_expr(ASRUtils::get_expr_size_expr(value));
 
         Location loc; loc.first = 1, loc.last = 1;
         ASRUtils::ASRBuilder builder(al, loc);
         Vec<ASR::dimension_t> realloc_dims;
+        size_t target_rank = ASRUtils::extract_n_dims_from_ttype(target_type);
         realloc_dims.reserve(al, target_rank);
         for( size_t i = 0; i < target_rank; i++ ) {
             ASR::dimension_t realloc_dim;
@@ -883,6 +981,17 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
                 al, loc, realloc_var, builder.i32(i + 1), int32, nullptr));
             realloc_dims.push_back(al, realloc_dim);
         }
+        ASR::expr_t* realloc_str_len {};
+        if(ASRUtils::is_character(*ASRUtils::expr_type(realloc_var))){
+            ASR::expr_t* len_value{}; // Compile-Time Length
+            int64_t len {};
+            if(ASRUtils::is_value_constant(ASR::down_cast<ASR::String_t>(
+                ASRUtils::extract_type(ASRUtils::expr_type(realloc_var)))->m_len), len) {
+                len_value = builder.i32(len);
+            }
+            realloc_str_len = ASRUtils::EXPR(ASR::make_StringLen_t(
+                al, loc, realloc_var, int32, len_value));
+        }
 
         Vec<ASR::alloc_arg_t> alloc_args; alloc_args.reserve(al, 1);
         ASR::alloc_arg_t alloc_arg;
@@ -890,14 +999,74 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         alloc_arg.m_a = target;
         alloc_arg.m_dims = realloc_dims.p;
         alloc_arg.n_dims = realloc_dims.size();
-        alloc_arg.m_len_expr = nullptr;
+        alloc_arg.m_len_expr = realloc_str_len;
         alloc_arg.m_type = nullptr;
+        alloc_arg.m_sym_subclass = nullptr;
         alloc_args.push_back(al, alloc_arg);
 
         pass_result.push_back(al, ASRUtils::STMT(ASR::make_ReAlloc_t(
-            al, loc, alloc_args.p, alloc_args.size())));
+                al, loc, alloc_args.p, alloc_args.size())));
     }
 
+    void visit_Allocate(const ASR::Allocate_t& x) {
+        ASR::Allocate_t& xx = const_cast<ASR::Allocate_t&>(x);
+        if (xx.m_source) {
+            pass_result.push_back(al, &(xx.base));
+            // Pushing assignment statements to source
+            for (size_t i = 0; i < x.n_args ; i++) {
+                if( !ASRUtils::is_array(
+                        ASRUtils::expr_type(x.m_args[i].m_a)) ) {
+                    continue;
+                }
+                ASR::stmt_t* assign = ASRUtils::STMT(ASRUtils::make_Assignment_t_util(
+                    al, xx.m_args[i].m_a->base.loc, xx.m_args[i].m_a, xx.m_source, nullptr, realloc_lhs, false));
+                ASR::Assignment_t* assignment_t = ASR::down_cast<ASR::Assignment_t>(assign);
+                Vec<ASR::expr_t**> fix_type_args;
+                fix_type_args.reserve(al, 2);
+                fix_type_args.push_back(al, const_cast<ASR::expr_t**>(&(assignment_t->m_target)));
+                fix_type_args.push_back(al, const_cast<ASR::expr_t**>(&(assignment_t->m_value)));
+
+                Vec<ASR::expr_t**> vars1;
+                vars1.reserve(al, 1);
+                if( ASRUtils::is_array(
+                        ASRUtils::expr_type(assignment_t->m_value)) ) {
+                    vars1.push_back(al, &(assignment_t->m_value));
+                }
+                vars1.push_back(al, &(assignment_t->m_target));
+
+                // TODO: Remove the following. Instead directly handle
+                // allocate with source in the backend.
+                if( xx.m_args[i].n_dims == 0 ) {
+                    Vec<ASR::expr_t**> vars2;
+                    vars2.reserve(al, 1);
+                    vars2.push_back(al, &(assignment_t->m_target));
+                    if( ASRUtils::is_array(
+                            ASRUtils::expr_type(assignment_t->m_value)) ) {
+                        vars2.push_back(al, &(assignment_t->m_value));
+                    }
+                    insert_realloc_for_target(
+                        xx.m_args[i].m_a, xx.m_source, vars2);
+                }
+                generate_loop(*assignment_t, vars1, fix_type_args, x.base.base.loc);
+            }
+        }
+    }
+
+    // Don't visit inside DebugCheckArrayBounds, it may contain ArrayConstant and result_expr will be nullptr
+    void visit_DebugCheckArrayBounds(const ASR::DebugCheckArrayBounds_t& x) {
+        (void)x;
+    }
+
+    bool is_looping_necessary_for_bitcast(ASR::expr_t* value) {
+        if (ASR::is_a<ASR::BitCast_t>(*value)) {
+            ASR::BitCast_t* bit_cast = ASR::down_cast<ASR::BitCast_t>(value);
+            return !ASRUtils::is_string_only(ASRUtils::expr_type(bit_cast->m_source));
+        } else {
+            return false;
+        }
+    }
+
+
     void visit_Assignment(const ASR::Assignment_t& x) {
         if (ASRUtils::is_simd_array(x.m_target)) {
             if( !(ASRUtils::is_allocatable(x.m_value) ||
@@ -924,6 +1093,12 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
             (ASRUtils::is_simd_array(xx.m_target) && ASRUtils::is_simd_array(xx.m_value)) ) {
             return ;
         }
+        bool is_target_assumed_rank = (ASR::is_a<ASR::ArrayPhysicalCast_t>(*xx.m_target) && 
+            ASR::down_cast<ASR::ArrayPhysicalCast_t>(xx.m_target)->m_old == ASR::array_physical_typeType::AssumedRankArray) 
+            || ASRUtils::is_assumed_rank_array(ASRUtils::expr_type(xx.m_target));
+        bool is_value_assumed_rank = (ASR::is_a<ASR::ArrayPhysicalCast_t>(*xx.m_value) && 
+            ASR::down_cast<ASR::ArrayPhysicalCast_t>(xx.m_value)->m_old == ASR::array_physical_typeType::AssumedRankArray)
+            || ASRUtils::is_assumed_rank_array(ASRUtils::expr_type(xx.m_value));
         xx.m_value = ASRUtils::get_past_array_broadcast(xx.m_value);
         const Location loc = x.base.base.loc;
 
@@ -953,12 +1128,20 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
         vars.reserve(al, 1);
         ArrayVarAddressCollector var_collector_target(al, vars);
         var_collector_target.current_expr = const_cast<ASR::expr_t**>(&(xx.m_target));
-        var_collector_target.call_replacer();
+        if (!is_target_assumed_rank) {
+            var_collector_target.call_replacer();
+        } else {
+            vars.push_back(al, const_cast<ASR::expr_t**>(&(xx.m_target)));
+        }
         ArrayVarAddressCollector var_collector_value(al, vars);
         var_collector_value.current_expr = const_cast<ASR::expr_t**>(&(xx.m_value));
-        var_collector_value.call_replacer();
+        if (!is_value_assumed_rank) {
+            var_collector_value.call_replacer();
+        } else {
+            vars.push_back(al, const_cast<ASR::expr_t**>(&(xx.m_value)));
+        }
 
-        if (vars.size() == 1 &&
+        if (vars.size() == 1 && !is_looping_necessary_for_bitcast(xx.m_value) && 
             ASRUtils::is_array(ASRUtils::expr_type(ASRUtils::get_past_array_broadcast(xx.m_value)))
         ) {
             return ;
@@ -968,16 +1151,45 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
             insert_realloc_for_target(xx.m_target, xx.m_value, vars);
         }
 
+        if (bounds_checking && 
+            ASRUtils::is_array(ASRUtils::expr_type(x.m_target)) &&
+            ASRUtils::is_array(ASRUtils::expr_type(x.m_value))) {
+            ASRUtils::ExprStmtDuplicator expr_duplicator(al);
+            ASR::expr_t* d_target = expr_duplicator.duplicate_expr(x.m_target);
+            ASR::expr_t* d_value = expr_duplicator.duplicate_expr(x.m_value);
+
+            Vec<ASR::expr_t*> vars;
+            vars.reserve(al, 1);
+
+            CollectComponentsFromElementalExpr cv(al, vars);
+            cv.visit_expr(*d_value);
+
+            if (debug_inserted.find(&x) == debug_inserted.end()) {
+                pass_result.push_back(al, ASRUtils::STMT(ASR::make_DebugCheckArrayBounds_t(al, x.base.base.loc, d_target, vars.p, vars.n, x.m_move_allocation)));
+                if (!x.m_move_allocation) {
+                    debug_inserted.insert(&x);
+                }
+            }
+        }
+
+        // Don't generate a loop for a move assignment
+        // The assignment should be handled in the backend
+        if (x.m_move_allocation) {
+            ASR::stmt_t* stmt = ASRUtils::STMT(ASRUtils::make_Assignment_t_util(al, loc, x.m_target, x.m_value, x.m_overloaded, x.m_realloc_lhs, x.m_move_allocation));
+            pass_result.push_back(al, stmt);
+            debug_inserted.insert(ASR::down_cast<ASR::Assignment_t>(stmt));
+            return;
+        }
+
         Vec<ASR::expr_t**> fix_type_args;
         fix_type_args.reserve(al, 2);
         fix_type_args.push_back(al, const_cast<ASR::expr_t**>(&(xx.m_target)));
         fix_type_args.push_back(al, const_cast<ASR::expr_t**>(&(xx.m_value)));
-
         generate_loop(x, vars, fix_type_args, loc);
     }
 
     void visit_SubroutineCall(const ASR::SubroutineCall_t& x) {
-        if( !PassUtils::is_elemental(x.m_name) ) {
+        if( !ASRUtils::is_elemental(x.m_name) ) {
             return ;
         }
         const Location loc = x.base.base.loc;
@@ -991,6 +1203,13 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
             }
         }
 
+        // If x.m_dt is an array, and the subroutine is elemental, we need to replace the m_dt with an array item
+        // and call the elemental function on each element of the x.m_dt array not on the whole array
+        if (x.m_dt && ASRUtils::is_array(ASRUtils::expr_type(x.m_dt))) {
+            ASR::SubroutineCall_t& xx = const_cast<ASR::SubroutineCall_t&>(x);
+            vars.push_back(al, &(xx.m_dt));
+        }
+
         if( vars.size() == 0 ) {
             return ;
         }
@@ -1035,7 +1254,7 @@ class ArrayOpVisitor: public ASR::CallReplacerOnExpressionsVisitor<ArrayOpVisito
 
 void pass_replace_array_op(Allocator &al, ASR::TranslationUnit_t &unit,
                            const LCompilers::PassOptions& pass_options) {
-    ArrayOpVisitor v(al, pass_options.realloc_lhs);
+    ArrayOpVisitor v(al, pass_options.realloc_lhs_arrays, pass_options.bounds_checking);
     v.call_replacer_on_value = false;
     v.visit_TranslationUnit(unit);
     PassUtils::UpdateDependenciesVisitor u(al);
